/*	$Id: imutil.c 2007/11/26 11:29:48 mallet $ */

/*
 * Copyright (c) 1997 LAAS/CNRS                       --  Thu Aug  3 2006
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include <math.h>
#include <float.h>

#include "libimages3d.h"

#define MAX(a,b)	((a)>(b)?(a):(b))
#define MIN(a,b)	((a)<(b)?(a):(b))

/**********************************************************************
 *
 * Dtermine les valeurs min et max de chaque coordonnes pour l'im3d
 *
 **********************************************************************/

int data_im3d_minmax (DATA_IM3D*  im3d)


{
  int         i;		/* Pour boucler sur les points */
  DATA_PT3D*  point;
  int         nbpoints, nbpoints_good;

  /* On vrifie que les points ont bien t dtermins */
  if (im3d->header.points != POINTS_CALCULES)
    return (ERR_IM3D_POINTS_NON_CALCULES);

  /* On rinitialise les bornes de l'ima3d (si jamais le calcul de borne avait
     dj t fait) */
  im3d->header.max_coord_1 = - FLT_MAX;
  im3d->header.min_coord_1 = FLT_MAX;
  im3d->header.max_coord_2 = - FLT_MAX;
  im3d->header.min_coord_2 = FLT_MAX;
  im3d->header.max_coord_3 = - FLT_MAX;
  im3d->header.min_coord_3 = FLT_MAX;
  nbpoints_good = 0;

  /* On boucle */
  nbpoints = im3d->header.nblig * im3d->header.nbcol;
  point = im3d->points_tab;
  for (i = 0; i < nbpoints; i++) {
    if (point->state == GOOD_PT3D) {
      im3d->header.max_coord_1 = MAX (im3d->header.max_coord_1, point->coord_1);
      im3d->header.max_coord_2 = MAX (im3d->header.max_coord_2, point->coord_2);
      im3d->header.max_coord_3 = MAX (im3d->header.max_coord_3, point->coord_3);
      im3d->header.min_coord_1 = MIN (im3d->header.min_coord_1, point->coord_1);
      im3d->header.min_coord_2 = MIN (im3d->header.min_coord_2, point->coord_2);
      im3d->header.min_coord_3 = MIN (im3d->header.min_coord_3, point->coord_3);
      nbpoints_good++ ;
    }
    point++;
  }

  if (nbpoints_good == 0) {
    im3d->header.max_coord_1 = 0;
    im3d->header.max_coord_2 = 0;
    im3d->header.max_coord_3 = 0;
    im3d->header.min_coord_1 = 0;
    im3d->header.min_coord_2 = 0;
    im3d->header.min_coord_3 = 0;
  }
  im3d->header.bornes = BORNES_CALCULEES;

  return (ERR_IM3D_OK);
}

/**********************************************************************
 *
 * Simple changement de repre
 *
 **********************************************************************/

int change_repere_data_im3d (DATA_IM3D*  im3d,
				   T3D*       transfo)


{
  int            i;		/* Pour boucler sur le tableau de points */
  int            nbcol, nblig;
  int            nb_points;
  DATA_PT3D*     point;		/* Un point de 'im3d' */
  float          x, y, z;	/* Ses coordonnes avant le chgt de repere */
  double         t0, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11;
				/* Les coefs de la matrice de la tr3d */
  T3D     tr_resultat;

  /* On vrifie qu'on a bien des points en coordonnes cartsiennes */
  if (im3d->header.points != POINTS_CALCULES)
    return (ERR_IM3D_POINTS_NON_CALCULES);
  if (im3d->header.coordonnees != COORDONNEES_CARTESIENNES)
    return (ERR_IM3D_TYPE_COORDONNEES);

  /* On 'ajoute' la transformation 'transfo' au champ tr_capteur_2_repere de
     l'image 3d  */
  if (im3d->header.repere == REPERE_CAPTEUR) {
    im3d->header.repere = REPERE_AUTRE;
    t3dCopyIn(&im3d->header.tr_capteur_2_repere, transfo);
  }
  else { /* Il faut composer la transfo dj existante avec la nouvelle */
    t3dComp(&im3d->header.tr_capteur_2_repere, transfo);
  }

  /* On rcupre les coefs de la matrice homogne */
  t3dCopyIn(&tr_resultat, transfo);
  t3dConvertTo(T3D_MATRIX, &tr_resultat);
  t0 = tr_resultat.matrix.matrix[0];
  t1 = tr_resultat.matrix.matrix[1];
  t2 = tr_resultat.matrix.matrix[2];
  t3 = tr_resultat.matrix.matrix[3];
  t4 = tr_resultat.matrix.matrix[4];
  t5 = tr_resultat.matrix.matrix[5];
  t6 = tr_resultat.matrix.matrix[6];
  t7 = tr_resultat.matrix.matrix[7];
  t8 = tr_resultat.matrix.matrix[8];
  t9 = tr_resultat.matrix.matrix[9];
  t10 = tr_resultat.matrix.matrix[10];
  t11 = tr_resultat.matrix.matrix[11];

  /* Et allez, on boucle */
  nblig = im3d->header.nblig;
  nbcol = im3d->header.nbcol;
  nb_points = nblig * nbcol;
  point = im3d->points_tab;

  for(i = 0; i < nb_points; i++) {
    if (point->state == GOOD_PT3D) {
      point->repere = REPERE_AUTRE;
      x = point->coord_1;
      y = point->coord_2;
      z = point->coord_3;
      point->coord_1 = (x * t0) + (y * t1) + (z * t2) + t3;
      point->coord_2 = (x * t4) + (y * t5) + (z * t6) + t7;
      point->coord_3 = (x * t8) + (y * t9) + (z * t10) + t11;
    }
    point++;
  }

  /* Si les variances sont dtermines dans l'image, on effectue aussi le
     changement de repre pour elles, avec la formule "Cov_inA = (A_2_B) *
     Cov_inB * (A_2_B)transpose". On fait a  la main... */
  if (im3d->header.covariances == COVARIANCES_CALCULEES) {
    double tempo[3][3];

    point = im3d->points_tab;
    for(i = 0; i < nb_points; i++) {
      if (point->state == GOOD_PT3D) {
	tempo[0][0] = *(point->covariances_tab) * t0
		      + *(point->covariances_tab + 1) * t1
		      + *(point->covariances_tab + 2) * t2;
	tempo[0][1] = *(point->covariances_tab) * t4
		      + *(point->covariances_tab + 1) * t5
		      + *(point->covariances_tab + 2) * t6;
	tempo[0][2] = *(point->covariances_tab) * t8
		      + *(point->covariances_tab + 1) * t9
		      + *(point->covariances_tab + 2) * t10;
	tempo[1][0] = *(point->covariances_tab + 1) * t0
		      + *(point->covariances_tab + 3) * t1
		      + *(point->covariances_tab + 4) * t2;
	tempo[1][1] = *(point->covariances_tab + 1) * t4
		      + *(point->covariances_tab + 3) * t5
		      + *(point->covariances_tab + 4) * t6;
	tempo[1][2] = *(point->covariances_tab + 1) * t8
		      + *(point->covariances_tab + 3) * t9
		      + *(point->covariances_tab + 4) * t10;
	tempo[2][0] = *(point->covariances_tab + 2) * t0
		      + *(point->covariances_tab + 4) * t1
		      + *(point->covariances_tab + 5) * t2;
	tempo[2][1] = *(point->covariances_tab + 2) * t4
		      + *(point->covariances_tab + 4) * t5
		      + *(point->covariances_tab + 5) * t6;
	tempo[2][2] = *(point->covariances_tab + 2) * t8
		      + *(point->covariances_tab + 4) * t9
		      + *(point->covariances_tab + 5) * t10;
	*(point->covariances_tab) = t0 * tempo[0][0]
				+ t1 * tempo[1][0]
				+ t2 * tempo[2][0];
	*(point->covariances_tab + 1) = t0 * tempo[0][1]
				    + t1 * tempo[1][1]
				    + t2 * tempo[2][1];
	*(point->covariances_tab + 2) = t0 * tempo[0][2]
				    + t1 * tempo[1][2]
				    + t2 * tempo[2][2];
	*(point->covariances_tab + 3) = t4 * tempo[0][1]
				    + t5 * tempo[1][1]
				    + t6 * tempo[2][1];
	*(point->covariances_tab + 4) = t4 * tempo[0][2]
				    + t5 * tempo[1][2]
				    + t6 * tempo[2][2];
	*(point->covariances_tab + 5) = t8 * tempo[0][2]
				    + t9 * tempo[1][2]
				    + t10 * tempo[2][2];
      }
      point++;
    }

  }


  /* Si jamais les bornes taient dj dtermines, on les recalcule */
  if (im3d->header.bornes == BORNES_CALCULEES) {

    im3d->header.max_coord_1 = - FLT_MAX;
    im3d->header.min_coord_1 = FLT_MAX;
    im3d->header.max_coord_2 = - FLT_MAX;
    im3d->header.min_coord_3 = FLT_MAX;
    im3d->header.max_coord_3 = - FLT_MAX;
    im3d->header.min_coord_3 = FLT_MAX;

    data_im3d_minmax (im3d);
  }

  return (ERR_IM3D_OK);
}

/**********************************************************************
 *
 * im3d_normales : dtermine les normales aux points des im3d
 *
 **********************************************************************/

int data_im3d_normales (DATA_IM3D*     im3d,
			      type_normales  type_calcul)


{
  int           nblig;
  int           nbcol;
  int           i, j;		/* Pour boucler */
  /* Le points courants et ses 8 voisins (h = haut, b = bas, d = droite, g =
     gauche) */
  DATA_PT3D*    pt_courant;
  DATA_PT3D*    pt_hg;
  DATA_PT3D*    pt_h;
  DATA_PT3D*    pt_hd;
  DATA_PT3D*    pt_d;
  DATA_PT3D*    pt_bd;
  DATA_PT3D*    pt_b;
  DATA_PT3D*    pt_bg;
  DATA_PT3D*    pt_g;

  /* les "drives" partielles */
  double        dx_i;
  double        dy_i;
  double        dz_i;
  double        dx_j;
  double        dy_j;
  double        dz_j;

  /* Les coordonnes du vecteur normal et sa norme */
  double        n_x;
  double        n_y;
  double        n_z;
  double        norme;

  int     retour_allocation; /* Pour tester les rsultat de l'allocation
				      mmoire */

  /* Validite du paramtre 'type_calcul' */
  if ((type_calcul != QUATRE_VOISINS) && (type_calcul != HUIT_VOISINS)) {
    printf ("Entre un 1 ou un 2 comme type de calcul des normales !\n");
    return (ERR_IM3D_TYPE_CALCUL_NORMAL);
  }

  /* On vrifie qu'on a bien des points en coordonnes cartsiennes */
  if (im3d->header.points != POINTS_CALCULES)
    return (ERR_IM3D_POINTS_NON_CALCULES);
  if (im3d->header.coordonnees != COORDONNEES_CARTESIENNES)
    return (ERR_IM3D_TYPE_COORDONNEES);


  /* Si les normales ne sont pas dj alloues, on le fait */
  if (im3d->header.normales == NIL_NORMALES) {
    retour_allocation = alloue_im3d_normales (im3d);
    if (retour_allocation != ERR_IM3D_OK)
      return (retour_allocation);
  }


  /* Let's go (il y a srement un moyen d'optimiser cette boucle) */
  nblig = im3d->header.nblig;
  nbcol = im3d->header.nbcol;

  switch (type_calcul) {
  case HUIT_VOISINS:
    for (i = 1; i < nblig - 1; i++) {
      /* On positionne les 9 pointeurs en haut de la colonne */
      pt_courant = im3d->points_tab + (i * nbcol) + 1;
      pt_g = pt_courant - 1;
      pt_hg = pt_g - nbcol;
      pt_h = pt_hg + 1;
      pt_hd = pt_hg + 1;
      pt_d = pt_hg + nbcol;
      pt_bd = pt_d + nbcol;
      pt_b = pt_bd - 1;
      pt_bg = pt_bd - 1;	/* J'ai mis les 8 points "circulairement" pour
				   n'avoir qu'une addition  chaque fois */

      for (j = 1; j < nbcol - 1; j++) {
	if ((pt_courant->state == GOOD_PT3D)
	    && (pt_hg->state == GOOD_PT3D)
	    && (pt_h->state == GOOD_PT3D)
	    && (pt_hd->state == GOOD_PT3D)
	    && (pt_d->state == GOOD_PT3D)
	    && (pt_bd->state == GOOD_PT3D)
	    && (pt_b->state == GOOD_PT3D)
	    && (pt_bg->state == GOOD_PT3D)
	    && (pt_g->state == GOOD_PT3D)) {
	  /* Calcul des drives partielles */
	  dx_i = (pt_bg->coord_1 - pt_hg->coord_1)
		 + (pt_b->coord_1 - pt_h->coord_1)
		 + (pt_bd->coord_1 - pt_hd->coord_1);
	  dy_i = (pt_bg->coord_2 - pt_hg->coord_2)
		 + (pt_b->coord_2 - pt_h->coord_2)
		 + (pt_bd->coord_2 - pt_hd->coord_2);
	  dz_i = (pt_bg->coord_3 - pt_hg->coord_3)
		 + (pt_b->coord_3 - pt_h->coord_3)
		 + (pt_bd->coord_3 - pt_hd->coord_3);
	  dx_j = (pt_hd->coord_1 - pt_hg->coord_1)
		 + (pt_d->coord_1 - pt_g->coord_1)
		 + (pt_bd->coord_1 - pt_bg->coord_1);
	  dy_j = (pt_hd->coord_2 - pt_hg->coord_2)
		 + (pt_d->coord_2 - pt_g->coord_2)
		 + (pt_bd->coord_2 - pt_bg->coord_2);
	  dz_j = (pt_hd->coord_3 - pt_hg->coord_3)
		 + (pt_d->coord_3 - pt_g->coord_3)
		 + (pt_bd->coord_3 - pt_bg->coord_3);

	  /* Dtermination du vecteur normal (produit vectoriel) */
	  n_x = (dy_i * dz_j) - (dy_j * dz_i);
	  n_y = (dx_j * dz_i) - (dx_i * dz_j);
	  n_z = (dx_i * dy_j) - (dx_j * dy_i);

	  /* Normalisation et mise  jour du flag du point */
	  norme = sqrt ( (n_x * n_x) + (n_y * n_y) + (n_z * n_z) );
	  *(pt_courant->normales_tab) = n_x / norme;
	  *(pt_courant->normales_tab + 1) = n_y / norme;
	  *(pt_courant->normales_tab + 2) = n_z / norme;
	  pt_courant->normales = NORMALES_CALCULEES;
	  pt_courant->nb_voisins_normales = HUIT_VOISINS;
	}

	/* On dcale tous les pointeurs d'un cran */
	pt_courant++;
	pt_hg++;
	pt_h++;
	pt_hd++;
	pt_d++;
	pt_bd++;
	pt_b++;
	pt_bg++;
	pt_g++;
      }
    }
    /* Mise  jour du flag 'nb_voisins_normales' de l'image */
    im3d->header.nb_voisins_normales = HUIT_VOISINS;
    break;

  case QUATRE_VOISINS:
    for (i = 1; i < nblig - 1; i++) {
      /* On positionne les 5 pointeurs en haut de la colonne */
      pt_courant = im3d->points_tab + (i * nbcol) + 1;
      pt_h = pt_courant - nbcol;
      pt_d = pt_courant + 1;
      pt_b = pt_courant + nbcol;
      pt_g = pt_courant - 1;

      for (j = 1; j < nbcol - 1; j++) {
	if ((pt_courant->state == GOOD_PT3D)
	    && (pt_h->state == GOOD_PT3D)
	    && (pt_d->state == GOOD_PT3D)
	    && (pt_b->state == GOOD_PT3D)
	    && (pt_g->state == GOOD_PT3D)) {
	  /* Calcul des drives partielles */
	  dx_i = (pt_b->coord_1 - pt_h->coord_1);
	  dy_i = (pt_b->coord_2 - pt_h->coord_2);
	  dz_i = (pt_b->coord_3 - pt_h->coord_3);
	  dx_j = (pt_d->coord_1 - pt_g->coord_1);
	  dy_j = (pt_d->coord_2 - pt_g->coord_2);
	  dz_j = (pt_d->coord_3 - pt_g->coord_3);

	  /* Dtermination du vecteur normal (produit vectoriel) */

	  n_x = (dy_i * dz_j) - (dy_j * dz_i);
	  n_y = (dx_j * dz_i) - (dx_i * dz_j);
	  n_z = (dx_i * dy_j) - (dx_j * dy_i);

	  /* Normalisation et mise  jour du flag du point */
	  norme = sqrt ( (n_x * n_x) + (n_y * n_y) + (n_z * n_z) );
	  *(pt_courant->normales_tab) = n_x / norme;
	  *(pt_courant->normales_tab + 1) = n_y / norme;
	  *(pt_courant->normales_tab + 2) = n_z / norme;
	  pt_courant->normales = NORMALES_CALCULEES;
	  pt_courant->nb_voisins_normales = QUATRE_VOISINS;
	}

	/* On dcale tous les pointeurs d'un cran */
	pt_courant++;
	pt_h++;
	pt_d++;
	pt_b++;
	pt_g++;
      }
    }
    /* Mise  jour du flag 'nb_voisins_normales' de l'image */
    im3d->header.nb_voisins_normales = QUATRE_VOISINS;
    break;

  default:
    return (ERR_IM3D_CAS_IMPOSSIBLE);
/*     break; */
  }

  /* Mise  jour du flag de l'image 3d */
  im3d->header.normales = NORMALES_CALCULEES;

  return (ERR_IM3D_OK);

}

/**********************************************************************
 *
 * extrait_sous_im3d : Extrait d'une ima3d la sous-image dfinie par les
 * quatre bornes passes en paramtre
 *
 *********************************************************************/


int extrait_sous_im3d (DATA_IM3D*  im3d,
			     DATA_IM3D*  sous_im3d,
			     int         i_min,
			     int         i_max,
			     int         j_min,
			     int         j_max)

{
  int          nblig;
  int          nbcol;
  int          i, j;

  int         delta_i;
  int         delta_j;		/* Taille de la sous-image */

  DATA_PT3D*  point;		/* un point de l'image */
  DATA_PT3D*  sous_point;	/* un point de la sous-image */

  int   retour_allocation = ERR_IM3D_OK;

  /* Vrification de la validit des bornes passes en paramtres */
  nblig = im3d->header.nblig;
  nbcol = im3d->header.nbcol;
  if ((i_min < 0) || (i_max >= nblig) || (j_min < 0) || (j_max >= nbcol)
      || (i_max <= i_min) || (j_max <= j_min)) {
    return (ERR_IM3D_TAILLE_SOUS_IMAGE);
  }

  /* Copie des champs gnraux de la sous-image, qui sont tous identiques 
     ceux de l'image originale (sauf 'sous-image') */
  sous_im3d->header.robot = im3d->header.robot;
  sous_im3d->header.capteur = im3d->header.capteur;
  sous_im3d->header.repere = im3d->header.repere;
  sous_im3d->header.coordonnees = im3d->header.coordonnees;
  sous_im3d->header.reduction = im3d->header.reduction;
  sous_im3d->header.sous_image = SOUS_IMAGE;

  for ( i = 0; i < 16; i++)
    sous_im3d->header.mat_intrinseque[i] = im3d->header.mat_intrinseque[i];

  /* On retrouve la taille de la sous-image  crer */
  delta_i = i_max - i_min;
  delta_j = j_max - j_min;

  /* Allocation du tableau de points de la sous_image */
  if (im3d->header.points != NIL_POINTS)
    retour_allocation = alloue_im3d_points (sous_im3d, delta_i, delta_j);
  if (retour_allocation != ERR_IM3D_OK)
    return (retour_allocation);
  sous_im3d->header.points = im3d->header.points;

  /* On positionne tous les champs de taille de l'image (attention, pour les
     min et max, ne pas oublier que l'image a peut-tre dj t extraite
     d'une autre image!) Rq: on fait a aprs la fct alloue_im3d_points, car
     elle a positionn ces champs (et ils ne collent pas). */
  sous_im3d->header.taux_reduction = im3d->header.taux_reduction;
  sous_im3d->header.nblig_initial = im3d->header.nblig_initial;
  sous_im3d->header.nbcol_initial = im3d->header.nbcol_initial;

  sous_im3d->header.i_min
    = im3d->header.i_min + (im3d->header.taux_reduction * i_min);
  sous_im3d->header.i_max
    = sous_im3d->header.i_min + (im3d->header.taux_reduction * delta_i);
  sous_im3d->header.j_min
    = im3d->header.j_min + (im3d->header.taux_reduction * j_min);
  sous_im3d->header.j_max
    = sous_im3d->header.j_min + (im3d->header.taux_reduction * delta_j);

  /* Allocation des autres tableaux de la sous_image */
  if (im3d->header.normales != NIL_NORMALES)
    retour_allocation = alloue_im3d_normales (sous_im3d);
  if (retour_allocation != ERR_IM3D_OK)
    return (retour_allocation);
  sous_im3d->header.normales = im3d->header.normales;
  sous_im3d->header.nb_voisins_normales = im3d->header.nb_voisins_normales;

  if (im3d->header.covariances != NIL_COVARIANCES)
    retour_allocation = alloue_im3d_covariances (sous_im3d);
  if (retour_allocation != ERR_IM3D_OK)
    return (retour_allocation);
  sous_im3d->header.covariances = im3d->header.covariances;

  if (im3d->header.local != NIL_INFOS_LOCALES)
    retour_allocation = alloue_im3d_infos_locales (sous_im3d, im3d->header.size_local);
  if (retour_allocation != ERR_IM3D_OK)
    return (retour_allocation);
  sous_im3d->header.local = im3d->header.local;

  /* Copie des points de la sous_image */
  /*   point = im3d->points_tab + (j_min * nblig) + i_min; */
  /*   sous_point = sous_im3d->points_tab; */

  for (i = 0; i < delta_i; i++) {
    for (j = 0; j < delta_j; j++) {

      point = data_im3d_pt (im3d, i + i_min, j + j_min);
      sous_point = data_im3d_pt (sous_im3d, i, j);

      copy_data_pt3d (point, sous_point, im3d->header.size_local);

/*       point++; */
/*       sous_point++; */
    }

    /* Attention l : il faut sauter une ligne et se retrouver au bon
       endroit... */
/*     point = point + nblig - delta_i; */
  }

  if (im3d->header.bornes == BORNES_CALCULEES)
    data_im3d_minmax (sous_im3d);

  sous_im3d->header.bornes = im3d->header.bornes;

  return (ERR_IM3D_OK);
}

/**********************************************************************
 *
 * Copie d'une DATA_IM3D (et/ou duplication, la difffrence est subtile,
 * voir le code).
 *
 *********************************************************************/

int
dup_im3d(DATA_IM3D *im3d, DATA_IM3D **copie)
{
   int s;
   s = cree_data_im3d(copie,
		      im3d->header.nblig,im3d->header.nbcol,
		      im3d->header.covariances != NIL_COVARIANCES,
		      im3d->header.normales != NIL_NORMALES,
		      (im3d->header.local != NIL_INFOS_LOCALES)*
		      im3d->header.size_local);
   if (s != ERR_IM3D_OK)
     return s;

   return copy_im3d(im3d, *copie);
}

int
copy_im3d(DATA_IM3D *im3d, DATA_IM3D *copie)
{
   return sous_echantillonne_im3d(im3d, 1, copie);
}

/**********************************************************************
 *
 * Simple sous-chantillonnage d'une DATA_IM3D
 *
 *********************************************************************/

int
sous_echantillonne_im3d(DATA_IM3D*  im3d,
			int          taux_reduction,
			DATA_IM3D*   im3d_reduite)
{
  int          i, j;

  int          nblig_reduite;
  int          nbcol_reduite;

  DATA_PT3D*  point;		/* un point de l'image */
  DATA_PT3D*  point_reduite;	/* un point de l'image sous-chantillone */

  int   retour_allocation = ERR_IM3D_OK;

  /* Vrification de la validit des bornes passes en paramtres */
  if (taux_reduction < 1) {
    return (ERR_IM3D_TAUX_REDUCTION);
  }

  /* Copie des champs gnraux de l'image rduite, qui sont tous identiques 
     ceux de l'image originale (sauf 'reduction') */
  im3d_reduite->header.robot = im3d->header.robot;
  im3d_reduite->header.capteur = im3d->header.capteur;
  im3d_reduite->header.repere = im3d->header.repere;
  im3d_reduite->header.coordonnees = im3d->header.coordonnees;
  im3d_reduite->header.sous_image = im3d->header.sous_image;

/*   if (taux_reduction == 1) */
/*     printf ("Sous-echantillonnage par 1. Idiot, non ?\n"); */
/*   else */
  im3d_reduite->header.reduction = REDUCTION;

  for ( i = 0; i < 16; i++)
    im3d_reduite->header.mat_intrinseque[i] = im3d->header.mat_intrinseque[i];

  /* Dtermination du taux de rduction de la nouvelle image */
  im3d_reduite->header.taux_reduction = (im3d->header.taux_reduction) * taux_reduction;

  /* Allocation des tableaux de la sous_image */
  if (im3d->header.points != NIL_POINTS &&
      im3d_reduite->header.points != NIL_POINTS) {

     /* On dtermine la taille de la sous-image  crer */
    nblig_reduite = im3d->header.nblig / taux_reduction;
    nbcol_reduite = im3d->header.nbcol / taux_reduction;

    retour_allocation
      = alloue_im3d_points(im3d_reduite, nblig_reduite, nbcol_reduite);
    if (retour_allocation != ERR_IM3D_OK)
      return (retour_allocation);

    im3d_reduite->header.points = im3d->header.points;
  }

  /* On positionne tous les champs de taille de l'image (attention, pour les
     min et max, ne pas oublier que l'image a peut-tre dj t extraite
     d'une autre image!) Rq: on fait a aprs la fct alloue_im3d_points, car
     elle a positionn ces champs (et ils ne collent pas). */
  im3d_reduite->header.nblig_initial = im3d->header.nblig_initial;
  im3d_reduite->header.nbcol_initial = im3d->header.nbcol_initial;

  im3d_reduite->header.i_min = im3d->header.i_min;
  im3d_reduite->header.i_max = im3d->header.i_min;
  im3d_reduite->header.j_min = im3d->header.j_min;
  im3d_reduite->header.j_max = im3d->header.j_max;


  if (im3d->header.normales != NIL_NORMALES &&
      im3d_reduite->header.normales != NIL_NORMALES) {
/*      retour_allocation = alloue_im3d_normales (im3d_reduite); */
/*      if (retour_allocation != ERR_IM3D_OK) */
/* 	return (retour_allocation); */
     im3d_reduite->header.normales = im3d->header.normales;
     im3d_reduite->header.nb_voisins_normales = im3d->header.nb_voisins_normales;
  } else if (im3d->header.normales == NIL_NORMALES) {

  }

  if (im3d->header.covariances != NIL_COVARIANCES &&
      im3d_reduite->header.covariances != NIL_COVARIANCES) {
/*      retour_allocation = alloue_im3d_covariances (im3d_reduite); */
/*      if (retour_allocation != ERR_IM3D_OK) */
/* 	return (retour_allocation); */
     im3d_reduite->header.covariances = im3d->header.covariances;
  }

  if (im3d->header.local != NIL_INFOS_LOCALES &&
      im3d_reduite->header.covariances != NIL_COVARIANCES) {

     if (im3d_reduite->header.size_local != im3d->header.size_local) {

	retour_allocation =
	alloue_im3d_infos_locales(im3d_reduite, im3d->header.size_local);
	if (retour_allocation != ERR_IM3D_OK)
	   return (retour_allocation);
	im3d_reduite->header.size_local = im3d->header.size_local;
     }
     im3d_reduite->header.local = im3d->header.local;
  }

  /* Copie des points de la sous_image */
  for (i = 0; i < nblig_reduite; i++) {
    for (j = 0; j < nbcol_reduite; j++) {

      point = data_im3d_pt (im3d, taux_reduction * i, taux_reduction * j);
      point_reduite = data_im3d_pt (im3d_reduite, i, j);

      copy_data_pt3d (point, point_reduite, im3d->header.size_local);

    }
  }

  im3d_reduite->header.points = POINTS_CALCULES;

  if (im3d->header.bornes == BORNES_CALCULEES)
    data_im3d_minmax (im3d_reduite);
  im3d_reduite->header.bornes = im3d->header.bornes;

  return (ERR_IM3D_OK);
}


/******************************************************************
 *
 * On met  0 les pixels de l'ima1 qui correspondent  un BAD_PT3D de l'im3d
 *
 ******************************************************************/

int filtre_ima1_im3d (DATA_IM3D*      im3d,
			    unsigned char*  q_image,
			    int             nblig,
			    int             nbcol)


{
  int             i;		/* Pour boucler */
  int             nb_points;	/* Nombre de pts des images  */
  DATA_PT3D*      point;	/* Un point de l'im3d */
  unsigned char*  pix;		/* Un pixel de l'ima1 */

  /* Validite des paramtres */
  /* On vrifie que l'im3d et l'image ont la mme taille */
  if ( (im3d->header.nblig != nblig) || (im3d->header.nbcol != nbcol) ) {
    return (ERR_IM3D_TAILLES_IMAGES);
  }

  /* On vrifie que les points sont dtermins */
  if (im3d->header.points != POINTS_CALCULES)
    return (ERR_IM3D_POINTS_NON_CALCULES);

  /* Boucle sur les points 3D */
  nb_points = nblig * nbcol;
  point = im3d->points_tab;
  pix = q_image;
  for (i = 0; i < nb_points; i++) {
    if (point->state == BAD_PT3D)
      *pix = 0;
    point++;
    pix++;
  }

  return (ERR_IM3D_OK);
}


/******************************************************************
 *
 * On met  BAD_PT3D les points 3D correspondant  un pixel de l'ima1
 *    avec niveau de luninance = 0 (typiquement, contours)
 *
 ******************************************************************/

int filtre_im3d_ima1 (DATA_IM3D*      im3d,
			    unsigned char*  q_image,
			    int             nblig,
			    int             nbcol)


{
  int             i;		/* Pour boucler */
  int             nb_points;	/* Nombre de pts des images  */
  DATA_PT3D*      point;	/* Un point de l'im3d */
  unsigned char*  pix;		/* Un pixel de l'ima1 */

  /* Validite des paramtres */
  /* On vrifie que l'im3d et l'image ont la mme taille */
  if ( (im3d->header.nblig != nblig) || (im3d->header.nbcol != nbcol) ) {
    return (ERR_IM3D_TAILLES_IMAGES);
  }

  /* On vrifie que les points sont dtermins */
  if (im3d->header.points != POINTS_CALCULES)
    return (ERR_IM3D_POINTS_NON_CALCULES);

  /* Boucle sur les points 3D */
  nb_points = nblig * nbcol;
  point = im3d->points_tab;
  pix = q_image;
  for (i = 0; i < nb_points; i++) {
    if (*pix == 0)
      point->state = BAD_PT3D;
    point++;
    pix++;
  }

  return (ERR_IM3D_OK);
}

/*******************************************************************
 *
 * limine d'une im3d les points dont une des coordonnes est suprieure (ou
 * Infrieure, a dpend du flag correspondant)  un seuil pass en paramtre
 *
 *******************************************************************/


int seuille_im3d (DATA_IM3D*  im3d,
			int     flag_1,
			double  seuil_1,
			int     flag_2,
			double  seuil_2,
			int     flag_3,
			double  seuil_3)

{
  int          i;		/* Pour boucler sur les points */
  int          nb_points;
  DATA_PT3D*   point;
  int          nb_points_seuilles_1 = 0;
  int          nb_points_seuilles_2 = 0;
  int          nb_points_seuilles_3 = 0;
  int          nb_good_points = 0;

  if (im3d->header.points != POINTS_CALCULES)
    return (ERR_IM3D_POINTS_NON_CALCULES);

  nb_points = im3d->header.nblig * im3d->header.nbcol;

  point = im3d->points_tab;

  for (i = 0; i < nb_points; i++) {

    if (point->state == GOOD_PT3D) {
      nb_good_points++;

      if ((flag_1 == -1) && (point->coord_1 < seuil_1)) {
	point->state = BAD_PT3D;
	nb_points_seuilles_1++;
      }
      else if ((flag_1 == 1) && (point->coord_1 > seuil_1)) {
	point->state = BAD_PT3D;
	nb_points_seuilles_1++;
      }

      if (point->state == GOOD_PT3D) {
	if ((flag_2 == -1) && (point->coord_2 < seuil_2)) {
	  point->state = BAD_PT3D;
	  nb_points_seuilles_2++;
	}
	else if ((flag_2 == 1) && (point->coord_2 > seuil_2)) {
	  point->state = BAD_PT3D;
	  nb_points_seuilles_2++;
	}
      }

      if (point->state == GOOD_PT3D) {
	if ((flag_3 == -1) && (point->coord_3 < seuil_3)) {
	  point->state = BAD_PT3D;
	  nb_points_seuilles_3++;
	}
	else if ((flag_3 == 1) && (point->coord_3 > seuil_3)) {
	  point->state = BAD_PT3D;
	  nb_points_seuilles_3++;
	}
      }
    }
    point++;
  }

  /* On redtermine les bornes de l'im3d si c'tait fait */
  /*  (bof) */


  printf ("Nb de points seuills :\n");
  printf ("\tsvt 1 : %d\n", nb_points_seuilles_1);
  printf ("\tsvt 2 : %d\n", nb_points_seuilles_2);
  printf ("\tsvt 3 : %d\n", nb_points_seuilles_3);
  printf ("Nb de points corrects :\n");
  printf ("\tavant seuillage: %d (%lf)\n",
	  nb_good_points, ((double) nb_good_points)/((double)nb_points));
  nb_good_points -= (nb_points_seuilles_1
		     + nb_points_seuilles_2
		     + nb_points_seuilles_3);
  printf ("\tapres seuillage: %d (%lf)\n",
	  nb_good_points, ((double)nb_good_points)/((double)nb_points));

  return (ERR_IM3D_OK);
}
