/*	$Id: ptutil.c 2007/11/23 19:06:59 mallet $ */

/*
 * Copyright (c) 1997 LAAS/CNRS                       --  Thu Aug  3 2006
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include <string.h>
#include <math.h>

#include "libimages3d.h"

/**********************************************************************
 *
 * Duplication du contenu d'un point 3d
 *
 *********************************************************************/

/* XXX mettre size_local dans la structure ? */

int
dup_data_pt3d(DATA_PT3D*  point,
	      DATA_PT3D**  copie,
	      int         size_local)
{
   int s;
   s = cree_data_pt3d(copie,
		      point->covariances != NIL_COVARIANCES,
		      point->normales != NIL_NORMALES,
		      (point->local != NIL_INFOS_LOCALES)*size_local);
   if (s != ERR_IM3D_OK)
      return s;

   return copy_data_pt3d(point, *copie, size_local);
}

int copy_data_pt3d (DATA_PT3D*  point,
			  DATA_PT3D*  copie,
			  int         size_local)

{
  int        i;

  copie->state = point->state;
  copie->repere = point->repere;
  copie->coordonnees = point->coordonnees;
/*   copie->normales = point->normales; */
  copie->nb_voisins_normales = point->nb_voisins_normales;
/*   copie->covariances = point->covariances; */

  copie->flag_perso = point->flag_perso;


  copie->coord_1 = point->coord_1;
  copie->coord_2 = point->coord_2;
  copie->coord_3 = point->coord_3;

  /* Rque: on fait la copie de ces champs, mme si c'est simplement
     allou et pas dtermin. C'est pas grave */

  if (point->normales != NIL_NORMALES &&
      copie->normales != NIL_NORMALES)
    for ( i = 0; i < IM3D_NORMALES_SIZE; i++ )
      *(copie->normales_tab + i) = *(point->normales_tab + i);

  if (point->covariances != NIL_COVARIANCES &&
      copie->covariances != NIL_COVARIANCES)
    for ( i = 0; i < IM3D_COVARIANCES_SIZE; i++ )
      *(copie->covariances_tab + i) = *(point->covariances_tab + i);

  if (point->local != NIL_INFOS_LOCALES &&
      copie->local != NIL_INFOS_LOCALES)
    memcpy ((void*) copie->local_tab, (void*) point->local_tab, size_local);

  return (ERR_IM3D_OK);
}

/***********************************************************************
 *
 * Calcule la normale (4 ou 8 voisins) du point dont les indices sont passs
 * en paramtre.
 *
 ***********************************************************************/
int calcule_normale_data_pt3d (DATA_IM3D*      im3d,
				     int             i,
				     int             j,
				     type_normales   type_calcul)

{
  int           nbcol;

  DATA_PT3D*    pt3d;

  /* Les 8 voisins de 'pt3d' (h = haut, b = bas, d = droite, g = gauche) */
  DATA_PT3D*    pt_hg;
  DATA_PT3D*    pt_h;
  DATA_PT3D*    pt_hd;
  DATA_PT3D*    pt_d;
  DATA_PT3D*    pt_bd;
  DATA_PT3D*    pt_b;
  DATA_PT3D*    pt_bg;
  DATA_PT3D*    pt_g;

  /* les "drives" partielles */
  double        dx_i;
  double        dy_i;
  double        dz_i;
  double        dx_j;
  double        dy_j;
  double        dz_j;

  /* Les coordonnes du vecteur normal et sa norme */
  double        n_x;
  double        n_y;
  double        n_z;
  double        norme;

  int     retour_allocation; /* Pour tester les rsultats de l'allocation
				      mmoire */

  /* On vrifie qu'on a bien des points en coordonnes cartsiennes */
  if (im3d->header.coordonnees != COORDONNEES_CARTESIENNES)
    return (ERR_IM3D_TYPE_COORDONNEES);

  /* On vrifie que le point n'est pas sur le bord de l'image */
  if ((i == 0)
      || (i == im3d->header.nblig - 1)
      || (j == 0)
      || (j == im3d->header.nbcol - 1)) {
    return (ERR_IM3D_INDICES_CALCUL_NORMALE_POINT);
  }

  /* Validite du paramtre 'type_calcul' */
  if ((type_calcul != QUATRE_VOISINS) && (type_calcul != HUIT_VOISINS)) {
    printf ("Entre un 1 ou un 2 comme type de calcul des normales !\n");
    return (ERR_IM3D_TYPE_CALCUL_NORMAL);
  }

  /* On rcupre le point */
  pt3d = im3d->points_tab + (i* im3d->header.nbcol) +j;

  /* Si les normales ne sont pas dj alloues, on le fait */
  if (pt3d->normales == NIL_NORMALES) {
    retour_allocation = alloue_pt3d_normales (pt3d);
    if (retour_allocation != ERR_IM3D_OK)
      return (retour_allocation);
  }


  /* Let's go (il y a srement un moyen d'optimiser cette boucle) */
  nbcol = im3d->header.nbcol;

  switch (type_calcul) {
  case QUATRE_VOISINS:
    pt_h = pt3d - nbcol;
    pt_d = pt3d + 1;
    pt_b = pt3d + nbcol;
    pt_g = pt3d - 1;

    if ((pt3d->state == GOOD_PT3D)
	&& (pt_h->state == GOOD_PT3D)
	&& (pt_d->state == GOOD_PT3D)
	&& (pt_b->state == GOOD_PT3D)
	&& (pt_g->state == GOOD_PT3D)) {
      /* Calcul des drives partielles */
      dx_i = (pt_b->coord_1 - pt_h->coord_1);
      dy_i = (pt_b->coord_2 - pt_h->coord_2);
      dz_i = (pt_b->coord_3 - pt_h->coord_3);
      dx_j = (pt_d->coord_1 - pt_g->coord_1);
      dy_j = (pt_d->coord_2 - pt_g->coord_2);
      dz_j = (pt_d->coord_3 - pt_g->coord_3);

      /* Dtermination du vecteur normal (produit vectoriel) */
      n_x = (dy_i * dz_j) - (dy_j * dz_i);
      n_y = (dx_j * dz_i) - (dx_i * dz_j);
      n_z = (dx_i * dy_j) - (dx_j * dy_i);

      /* Normalisation et mise  jour du flag du point */
      norme = sqrt ( (n_x * n_x) + (n_y * n_y) + (n_z * n_z) );
      *(pt3d->normales_tab) = n_x / norme;
      *(pt3d->normales_tab + 1) = n_y / norme;
      *(pt3d->normales_tab + 2) = n_z / norme;
      pt3d->normales = NORMALES_CALCULEES;
      pt3d->nb_voisins_normales = type_calcul;
    }
    break;
  case HUIT_VOISINS:
    pt_g = pt3d - 1;
    pt_hg = pt_g - nbcol;
    pt_h = pt_hg + 1;
    pt_hd = pt_hg + 1;
    pt_d = pt_hg + nbcol;
    pt_bd = pt_d + nbcol;
    pt_b = pt_bd - 1;
    pt_bg = pt_bd - 1;		/* J'ai mis les 8 points "circulairement" pour
				 n'avoir qu'une addition  chaque fois */

    if ((pt3d->state == GOOD_PT3D)
	&& (pt_hg->state == GOOD_PT3D)
	&& (pt_h->state == GOOD_PT3D)
	&& (pt_hd->state == GOOD_PT3D)
	&& (pt_d->state == GOOD_PT3D)
	&& (pt_bd->state == GOOD_PT3D)
	&& (pt_b->state == GOOD_PT3D)
	&& (pt_bg->state == GOOD_PT3D)
	&& (pt_g->state == GOOD_PT3D)) {
      /* Calcul des drives partielles */
      dx_i = (pt_bg->coord_1 - pt_hg->coord_1)
	     + (pt_b->coord_1 - pt_h->coord_1)
	     + (pt_bd->coord_1 - pt_hd->coord_1);
      dy_i = (pt_bg->coord_2 - pt_hg->coord_2)
	     + (pt_b->coord_2 - pt_h->coord_2)
	     + (pt_bd->coord_2 - pt_hd->coord_2);
      dz_i = (pt_bg->coord_3 - pt_hg->coord_3)
	     + (pt_b->coord_3 - pt_h->coord_3)
	     + (pt_bd->coord_3 - pt_hd->coord_3);
      dx_j = (pt_hd->coord_1 - pt_hg->coord_1)
	     + (pt_d->coord_1 - pt_g->coord_1)
	     + (pt_bd->coord_1 - pt_bg->coord_1);
      dy_j = (pt_hd->coord_2 - pt_hg->coord_2)
	     + (pt_d->coord_2 - pt_g->coord_2)
	     + (pt_bd->coord_2 - pt_bg->coord_2);
      dz_j = (pt_hd->coord_3 - pt_hg->coord_3)
	     + (pt_d->coord_3 - pt_g->coord_3)
	     + (pt_bd->coord_3 - pt_bg->coord_3);

      /* Dtermination du vecteur normal (produit vectoriel) */

      n_x = (dy_i * dz_j) - (dy_j * dz_i);
      n_y = (dx_j * dz_i) - (dx_i * dz_j);
      n_z = (dx_i * dy_j) - (dx_j * dy_i);

      /* Normalisation et mise  jour du flag du point */
      norme = sqrt ( (n_x * n_x) + (n_y * n_y) + (n_z * n_z) );
      *(pt3d->normales_tab) = n_x / norme;
      *(pt3d->normales_tab + 1) = n_y / norme;
      *(pt3d->normales_tab + 2) = n_z / norme;
      pt3d->normales = NORMALES_CALCULEES;
      pt3d->nb_voisins_normales = type_calcul;
    }

    break;
  default:
    return (ERR_IM3D_CAS_IMPOSSIBLE);
/*     break; */
  }

  return (ERR_IM3D_OK);
}

/**********************************************************************
 *
 * Changement de repre d'un point 3D
 *
 *********************************************************************/

int change_repere_data_pt3d (DATA_PT3D*  pt3d,
				   T3D*       transfo)

{
  T3D matrix;
  float    x, y, z;

  if (pt3d->coordonnees != COORDONNEES_CARTESIENNES)
    return (ERR_IM3D_TYPE_COORDONNEES);

  t3dCopyIn(&matrix, transfo);
  t3dConvertTo(T3D_MATRIX, &matrix);

  if (pt3d->state == GOOD_PT3D) {
    pt3d->repere = REPERE_AUTRE;
    x = pt3d->coord_1;
    y = pt3d->coord_2;
    z = pt3d->coord_3;
    pt3d->coord_1 = (x * matrix.matrix.matrix[0])
		     + (y * matrix.matrix.matrix[1])
		     + (z * matrix.matrix.matrix[2])
		     + matrix.matrix.matrix[3];
    pt3d->coord_2 = (x * matrix.matrix.matrix[4])
		     + (y * matrix.matrix.matrix[5])
		     + (z * matrix.matrix.matrix[6])
		     + matrix.matrix.matrix[7];
    pt3d->coord_3 = (x * matrix.matrix.matrix[8])
		     + (y * matrix.matrix.matrix[9])
		     + (z * matrix.matrix.matrix[10])
		     + matrix.matrix.matrix[11];
  }

  return (ERR_IM3D_OK);
}

